/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_THREAD_BASE
#define SENDS_THREAD_BASE
#include <pthread.h>


namespace thread {

    /**  Thread_base is a base class for threads. There is a single
      *  entry point for the thread.
      */
    class thread_base {
    public:
	/**  Construct the thread base class.
	  */
	thread_base(void);

	/**  Destroy the thread_bas class.
	  */
	virtual ~thread_base(void);

	/**  Test whether the thread is currently executing.
	  */
	bool busy(void) const;

	/**  Send a signal to this thread.
	  */
	virtual int kill_thread(int sig);

	/**  Test the thread run-enable flag.
	 */
	bool run(void) const;

	/**  Start the thread. The run flag is set and the busy flag is set 
	  *  if the thread is successfully started.
	  */
	virtual int start_thread(void);

	/**  Clear the run-enable flag.
	  */
	virtual int stop_thread(void);

	/**  Method to be executed when the thread is started.
	  */
	virtual void* thread_entry(void) = 0;
    private:
	/**  thread_base stub method to handle thread status flags.
	  */
	void* thread_stub(void);
	static void* launch(void* p);
    private:
	volatile bool mRun;
	volatile bool mBusy;
	pthread_t     mThreadID;
    };

    //==================================  Inline methods
    inline bool
    thread_base::busy(void) const {
	return mBusy;
    }

    inline bool
    thread_base::run(void) const {
	return mRun;
    }
}

#endif // !defined(SENDS_THREAD_BASE)
