#include <time.h>
#include <string>
#include <iostream>
#include "goption.hh"
#include "TimingReader.hh"
#include "lsmp_con.hh"

using namespace std;
using namespace xml;


TimingReader::TimingReader (int argc, char* argv[])
: mOpen (false), mCons (0), mBufid (0), mTerm (0), mHelp (false), mDebug (false)
{
   init (argc, argv);
}

TimingReader::~TimingReader()
{
   halt();
}

void TimingReader::read (void)
{
   //-----------------------------  Read in a data buffer
   const char *buffer =  mCons->get_buffer();
   if (!buffer) return;
   int len = mCons->getLength();
   if (mDebug)
       cerr << "Buffer number " << mBufid <<  "    Length " 
	      << mCons->getLength() << endl;

   //-----------------------------  Read timing data from shared memory buffer
   mfo_tree_list mfolist;
   xsilHandlerQueryMfoTree mfoQ (mfolist);
   xsilParser parser;
   parser.AddHandler (mfoQ);
   parser.Parse (buffer, len);
   
   //-----------------------------  Dump the timing data to stdout as XML
   work (mfolist); 
   
   //-----------------------------  Delete the frame/release the buffer.
   mBufid++;
   mCons->free_buffer();
   return;
}

//______________________________________________________________________________
//    init - set up the internal environment for pushing data.
//       * Create a shared memory partition for data distribution.
//       * Make a connection to the CDS Network Data Server. Register 
//         notification of all frames.
//       * Set up the termination signal handling.
//
bool TimingReader::init (int argc, char **argv)
{
   string	ShMemName (kSMDEFNAME); // shared memory partition name
   int 		ShMemSize = kSMDEFSIZE;	// partition size
   int 		ShMemNum = kSMDEFNUM; // number of partitions

   if (mOpen) return true;
   
   // parse command line parameters
   gdsbase::option_string opt (argc, argv, "m:s:n:dh");
   mDebug = opt.opt ('d');
   if (!opt.getOpt ('m', ShMemName)) {
      ShMemName = kSMDEFNAME;
   }
   if (!opt.getOpt ('s', ShMemSize)) {
      ShMemSize = kSMDEFSIZE;
   }
   if (!opt.getOpt ('n', ShMemNum)) {
      ShMemNum = kSMDEFNUM;
   }
   if (opt.help()) {
      mHelp = true;
      return false;
   }

   //---------------------------------  Create a shared memory consumer
   mCons = new LSMP_CON (ShMemName.c_str(), ShMemNum, ShMemSize);

   if (!mCons) return false;
   if (!mCons->valid()) {
       cerr << "Can't open partition " << ShMemName << ": " 
		 << mCons->Error() << endl;
       return false;
   }

   //---------------------------------  Set up the signal flags
   mTerm = new SigFlag(SIGINT);
   mTerm->add(SIGTERM);
   mTerm->add(SIGHUP);

   //---------------------------------  Done, quit.
   mOpen = true;
   return true;
}

//______________________________________________________________________________
//    halt - set down.
//       * Release the shared memory partition.
//       * Cancel the termination signal handling.
//
void TimingReader::halt (void)
{
   //---------------------------------  Disable the signal
   delete mTerm;
   mTerm = 0;

   //---------------------------------  Delete the partition
   delete mCons;
   mCons = 0;
   mOpen = false;
}
