#ifndef TimingTypes_HH
#define TimingTypes_HH

#include <inttypes.h>
#include <string>
#include <list>
#include <iosfwd>
#include "Time.hh"
#include "xml/Xsil.hh"

/// Default shared memory size
const char* const kSMDEFNAME = "TimingSystem";
/// Default shared memory size
const int kSMDEFSIZE = 0x100000;
/// Default shared memory buffer number
const int kSMDEFNUM = 8;

// Xsil constants
const char* const mfoName = "Module";
const char* const mfoGPS = "GPS";
const char* const mfoAddr = "Address";
const char* const mfoAddrNtuple = "AddressNtuple";
const char* const mfoBoard = "Board";
const char* const mfoBoardHex = "BoardHex";
const char* const mfoSerial = "Serial";
const char* const mfoProgram = "Program";
const char* const mfoProgramHex = "ProgramHex";
const char* const mfoRevision = "Revision";
const char* const mfoConfiguration = "Configuration";
const char* const mfoConfigurationHex = "ConfigurationHex";
const char* const mfoMaster = "IsMaster";
const char* const mfoFanout = "HasFanout";
const char* const mfoPorts = "Ports";
const char* const mfoUseExtPPS = "UseExtPPS";
const char* const mfoUseGPSPPS = "UseGPSPPS";
const char* const mfoUseUplinkPPS = "UseUplinkPPS";
const char* const mfoOCXOLocked = "OCXOLocked";
const char* const mfoGPSLocked = "GPSLocked";
const char* const mfoGPSPresent = "HasGPS";
const char* const mfoOCXOPresent = "HasOCXO";
const char* const mfoExtPPSPresent = "HasExtPPS";
const char* const mfoStatus = "Status";
const char* const mfoStatusHex = "StatusHex";
const char* const mfoLOS = "LOS";
const char* const mfoUp = "Up";
const char* const mfoErrC = "ErrorCount";
const char* const mfoDIP = "DIP";
const char* const mfoVCXOControl = "VCXOControl";
const char* const mfoOCXOControl = "OCXOControl";
const char* const mfoOCXOError = "OCXOErorr";
const char* const mfoExtPPSDelay = "ExtPPSDelay";
const char* const mfoUplinkDelay = "UplinkDelay";
const char* const mfoGPSDelay = "GPSDelay";
const char* const mfoFanoutUp = "FanoutUp";
const char* const mfoFanoutLOS = "FanoutLOS";
const char* const mfoFanoutMissing = "FanoutMissingDelay";
const char* const mfoFanoutDelayError = "FanoutDelayError";
const char* const mfoFlags = "Flags";
const char* const mfoCRCErrC = "CRCErrorCount";
const char* const mfoGPSErr = "GPSError";
const char* const mfoGPSErrC = "GPSErrorCount";
const char* const mfoSlaveOK = "SlaveCRCOK";
const char* const mfoGPSStatus = "GPSStatus";
const char* const mfoGPSLatitude = "Latitude";
const char* const mfoGPSLongitude = "Longitude";
const char* const mfoGPSHeight = "Height";
const char* const mfoGPSSpeed3D = "Speed3D";
const char* const mfoGPSSpeed2D = "Speed2D";
const char* const mfoGPSHeading = "Heading";
const char* const mfoGPSDOP = "GPSDOP";
const char* const mfoGPSVisSatellites = "GPSSatellitesVisible";
const char* const mfoGPSTrackSatellites = "GPSSatellitesTracking";
const char* const mfoGPSReceiver = "GPSReceiverStatus";
const char* const mfoGPSReceiverHex = "GPSReceiverStatusHex";
const char* const mfoGPSFix = "GPSFix";
const char* const mfoGPSNarrowBand = "GPSNarrowBand";
const char* const mfoGPSAntenna = "GPSAntennaOK";
const char* const mfoGPSSerial = "GPSSerial";
const char* const mfoExtended = "Extended";
const char* const mfoFOCRCErrorC = "CRCErrorCountFanout";
const char* const mfoCRC = "CRC";

const char* const slaveGPS = "GPS";
const char* const slaveAddr = "Address";
const char* const slaveAddrNtuple = "Address";
const char* const slaveBoard = "Board";
const char* const slaveBoardHex = "BoardHex";
const char* const slaveSerial = "Serial";
const char* const slaveProgram = "Program";
const char* const slaveProgramHex = "ProgramHex";
const char* const slaveRevision = "Revision";
const char* const slaveStatus = "Status";
const char* const slaveStatusHex = "StatusHex";
const char* const slaveLOS = "LOS";
const char* const slaveUp = "Up";
const char* const slaveErrC = "ErrorCount";
const char* const slaveDIP = "DIP";
const char* const slaveVCXOControl = "VCXOControl";
const char* const slaveConfiguration = "Configuration";
const char* const slaveConfigurationHex = "ConfigurationHex";
const char* const slaveCRCErrC = "CRCErrorCount";
const char* const slaveExtended = "Extended";
const char* const slaveCRCOK = "CRCOK";
const char* const slaveType = "Type";

const char* const slaveComparatorExtPPSPresent = "HasExtPPS";
const char* const slaveComparatorExtPPSDelay = "ExtPPSDelay";

const char* const slaveXOLockingPresent = "HasOCXO";
const char* const slaveXOLockingLocked = "OCXOLocked";
const char* const slaveXOLockingError = "OCXOError";
const char* const slaveXOLockingControl = "OCXOControl";
const char* const slaveXOLockingFreqSet = "SetFrequency";
const char* const slaveXOLockingFreqMeas= "OCXOFrequency";

const char* const portLOS = "LOS";
const char* const portUp = "Up";
const char* const portDelayError = "DelayError";
const char* const portMissing = "MissingDelay";
const char* const portErrCount = "ErrorCount";
const char* const portDelay = "Delay";
const char* const portAdvance = "Advance";
const char* const portUsedAdvance = "UsedAdvance";


/// Denotes a generic slave
const int kSlaveGeneric = 0;
/// Denotes that slave is a fanout module
const int kSlaveFanout = 1;
/// ID of fanout slave (DCC number of schematics)
const uint32_t kSlaveFanoutId = 0x070011;
/// Denotes a DuoTone slave
const int kSlaveDuoTone = 2;
const uint32_t kSlaveDuoToneId = 0x080335;
/// Denotes a timing comparator slave
const int kSlaveComparator = 3;
const uint32_t kSlaveComparatorId = 0x070568;
/// Denotes a oscillator locking slave
const int kSlaveXOLocking = 4;
const uint32_t kSlaveXOLockingId = 0x080665;


/// Fanout port structure
struct mfo_fanout_t {
   /// Measured Delay
   int32_t 		mDelayMeasured;
   /// Delay Control value and status
   uint32_t 		mDelayControl;
   
   /// Constructor
   mfo_fanout_t();
   /// Returns the measured fiber round-trip delay
   double measured_delay() const;
   /// Returns the requested time advance of the outgoing fiber
   double advance() const;
   /// Returns the used time advance of the outgoing fiber
   double used_advance() const;
   /// True if interface is up and running
   bool up() const;
   /// True if loss-of-signal detected
   bool los() const;
   /// True if a delay error was detected
   bool delayerr() const;
   /// True if the delay measurement was missing
   bool missing() const;
   /// Returns the error counter (will overflow)
   int errcount() const;

   // set the measured fiber round-trip delay
   bool set_measured_delay (double t);   
   /// Set the time advance of the outgoing fiber
   bool set_advance (double t);
   /// Set if interface is up and running
   bool set_up (bool b);
   /// Set if loss-of-signal detected
   bool set_los (bool b);
   /// Set if a delay error was detected
   bool set_delayerr (bool b);
   /// Set if the delay measurement was missing
   bool set_missing (bool b);
   /// Set the error counter
   bool set_errcount (int c);
};


/// Basic slave module structure
struct slave_basic_t {
   /// Slave board id
   uint32_t 		mBoardId;
   /// Slave board serial number
   uint32_t 		mBoardSerial;
   /// Software id
   uint32_t 		mSoftwareId;
   /// Software revision
   uint32_t 		mSoftwareRev;
   /// GPS time
   uint32_t 		mGPS;
   /// Slave address
   uint32_t 		mAddr;
   /// SLave status
   uint32_t 		mStatus;
   /// Slave configuration
   uint32_t 		mConf;

   /// Constructor
   slave_basic_t();
   /// Is a fanout module?
   bool is_fanout() const;
   /// Returns the slave  board revision (A becomes 0, F becomes 5, 0 becomes 6, etc.)
   int revision() const;
   /// Returns the slave daughter board type
   int daughter() const;
   /// Returns the slave daughter board revision (A becomes 0, F becomes 5, 0 becomes 6, etc.)
   int daughter_revision() const;
   
   /// VCXO control voltage
   double vcxo() const;
   /// DIP settings (1 through 8) for board; 9 and 10 for daughter board
   bool sw (int num) const;
   /// error counter
   int errcount() const;
   /// uplink is up and running
   bool up() const;
   /// loss-of-signal of uplink
   bool los() const;
   /// CRC error count
   int crc_errcount() const;

};


/// Generic slave module structure
struct slave_t : public slave_basic_t {
   /// Padding
   uint32_t		mExtended[24];
   
   /// Constructor
   slave_t();
   /// Compute CRC (should be zero for OK)
   uint32_t CRC() const;
};


/// Fanout slave
struct slave_fanout_t : public slave_t {

   /// Constructor
   slave_fanout_t() {};
};

/// DuoTone slave
struct slave_duotone_t : public slave_t {

   /// Constructor
   slave_duotone_t() {};
};

/// Timing comparator slave
struct slave_comparator_t : public slave_t {

   /// Constructor
   slave_comparator_t() {};
   /// external 1PPS present on input i (0 to 6)
   bool has_ext (int i) const;
   /// Returns the measured delay on input i (0 to 6)
   double ext_delay (int i) const;
};

/// XO locking slave
struct slave_xolocking_t : public slave_t {

   /// Constructor
   slave_xolocking_t() {};
   /// nominal frequency (set by push wheels)
   double freq_set() const;
   /// measured ocxo frequency
   double freq_meas() const;
   /// OCXO present on PLL
   bool has_ocxo() const;
   /// OCXO is locked
   bool ocxo_locked() const;
   /// ocxo control voltage
   double ocxo() const;
   /// phase delay of digital PLL (in sec)
   double ocxo_delay() const;
};


/// master/fanout module structure
struct mfo_t {
   /// master/fanout board id
   uint32_t 		mBoardId;
   /// master/fanout serial number
   uint32_t 		mBoardSerial;
   /// master/fanout software id
   uint32_t 		mSoftwareId;
   /// master/fanout software revision
   uint32_t 		mSoftwareRev;
   /// GPS time
   uint32_t 		mGPS;
   /// master/fanout address
   uint32_t 		mAddr;
   /// master/fanout status
   uint32_t 		mStatus;
   /// master/fanout configuration
   uint32_t 		mConf;
   /// OCXO control value
   uint32_t 		mOCXOControl;
   /// OCXO error value
   int32_t 		mOCXOError;
   /// Measured delay of 1PPS form uplink receiver
   int32_t 		mDelayUplink;
   /// Measured delay from external 1PPS
   int32_t 		mDelayExt;
   /// Measured delay from GPS
   int32_t 		mDelayGPS;
   /// Up and LOS flags from fanout ports
   uint32_t 		mFanoutUpLos;
   /// Missing and Delay error flags from fanout ports
   uint32_t 		mFanoutMissDerr;
   /// GPS errors and uplink CRC errors
   uint32_t 		mFlags;
   /// Fanout port structure
   mfo_fanout_t         mFO[16];
   /// GPS status message
   uint32_t		mGPSStatus[8];
   /// CRC error count of fanouts 16 * 8 bits and 4 dword padding
   uint32_t		mExtended[8];
   /// information from the slaves connected to the fanout ports
   slave_t		mSlave[16];
   /// CRC  value
   uint32_t		mCRC;
   
   /// Constructor
   mfo_t();
   /// Returns the board revision (A becomes 0, F becomes 5, 0 becomes 6, etc.)
   int revision() const;
   /// VCXO control voltage
   double vcxo() const;
   /// DIP settings (1 through 8) for board; 9 and 10 for daughter board
   bool sw (int num) const;
   /// error counter
   int errcount() const;
   /// uplink is up and running
   bool up() const;
   /// loss-of-signal of uplink
   bool los() const;
   /// OCXO present
   bool has_ocxo() const;
   /// external 1PPS present
   bool has_ext() const;
   /// GPS module is present
   bool has_gps() const;
   /// GPS is locked to satellites
   bool gps_locked() const;
   /// OCXO is locked
   bool ocxo_locked() const;
   /// external 1PPS is used to lock OCXO
   bool use_ext() const;
   /// 1PPS from GPS module is used
   bool use_gps() const;
   /// 1PPS from fiber uplink is used
   bool use_uplink() const;
   /// this is a master
   bool is_master() const;
   /// do we have fanout ports?
   bool has_fanout() const;
   /// number of fanout ports
   int fanout_ports() const;
   /// CRC error counters (-1 for uplink, port>=0 for fanout ports)
   int crc_err (int port) const;
   /// GPS error flag
   bool gps_err () const;
   /// GPS error count
   int gps_errcount () const;
   
   /// ocxo control voltage
   double ocxo() const;
   /// phase delay of digital PLL (in sec)
   double ocxo_delay() const;
   /// measured delay of uplink 1PPS
   double uplink_delay() const;
   /// measured delay of GPS 1PPS
   double gps_delay() const;
   /// measured delay of external 1PPS
   double ext_delay() const;
   /// set ocxo control voltage
   bool set_ocxo(double v);
   /// set phase delay of digital PLL (in sec)
   bool set_ocxo_delay(double t);
   /// set measured delay of uplink 1PPS
   bool set_uplink_delay(double t);
   /// set measured delay of GPS 1PPS
   bool set_gps_delay(double t);
   /// set measured delay of external 1PPS
   bool set_ext_delay(double t);
   
   /// fanout LOS
   bool fanout_los (int node) const;
   /// fanout Up
   bool fanout_up (int node) const;
   /// fanout missing delay
   bool fanout_missing (int node) const;
   /// fanout delay error
   bool fanout_derror (int node) const;
   /// set fanout LOS
   bool set_fanout_los (int node, bool val);
   /// set fanout Up
   bool set_fanout_up (int node, bool val);
   /// set fanout missing delay
   bool set_fanout_missing (int node, bool val);
   /// set fanout delay error
   bool set_fanout_derror (int node, bool val);
   
   /// GPS latitude in degrees (-90 to +90)
   double gps_latitude() const;
   /// GPS longitude in degrees (-180 to +180)
   double gps_longitude() const;
   /// GPS height in meters (-1000 to +18000)
   double gps_height() const;
   /// GPS 3D speed in m/s (0.0 to +514 m/s)
   double gps_3d_speed() const;
   /// GPS 2D speed in m/s (0.0 to +514 m/s)
   double gps_2d_speed() const;
   /// GPS heading in degree (0 to 360)
   double gps_heading() const;
   /// GPS Dilution of Precision (0 to 99.9)
   double gps_dop() const;
   /// GPS number of visible satellites (0 to 99.9)
   int gps_vis_satellites() const;
   /// GPS number of tracked satellites (0 to 99.9)
   int gps_track_satellites() const;
   /** GPS receiver status word
       The GPS message receiver decodes the "Hb" position/timing messages sent 
       by the GPS module. The information is made available as a status string. 
       Its format is 'aaaaoooohhhhVVvvhhddntssrrvvvvvv'.

       aaaa = latitude in mas -324,000,000..324,000,000 (-90..+90)
       oooo = longitude in mas -648,000,000..648,000,000 (-180..+180)
       hhhh = GPS height in cm -100,000..+1,800,000 (-1000..+18,000m)
       VV = 3D speed in cm/s 0..51400 (0.0 to 514 m/s)
       vv = 2D speed in cm/s 0..51400 (0.0 to 514 m/s)
       hh = 2D heading 0..3599 tenths of degrees (0.0 to 359.9)
       dd =current DOP 0..999 (0.0 to 99.9 DOP)
       n = number of visible satellites 0..12
       t = number of tracked satellites 0..12
       ss = receiver status
               Bits 15-13 (msb): 
                       111 = 3D Fix
                       110 = 2D Fix
                       101 = Propagate Mode
                       100 = Position Hold
                       011 = Acquiring Satellites
                       010 = Bad Geometry
                       001 = Reserved
                       000 = Reserved
               Bits 12-11: Reserved
               Bit 10: Receiver in narrow-band tracking mode (M12M only)
               Bit 9: Fast Acquisition Position
               Bit 8: Filter Reset To Raw GPS Solution
               Bit 7: Cold Start (no almanac, almanac out of date)
               Bit 6: Differential Fix
               Bit 5: Position Lock
               Bit 4: Autosurvey Mode
               Bit 3: Insufficient Visible Satellites
               Bits 2-1: Antenna Sense 00 = OK
                       01 = Overcurrent
                       10 = Undercurrent
                       11 = No bias voltage
               Bit 0: Code Location 
                       0 = EXTERNAL
                       1 = INTERNAL
       rr = reserved
       vvvvvv = ID tag 6 characters (0x20 to 0x7e), serial number by default
   */
   int gps_status() const;
   /// GPS fix
   std::string gps_fix() const;
   /// GPS narrow band mode
   bool gps_narrowband() const;
   /// GPS antenna ok?
   bool gps_antenna() const;
   /// GPS serial number
   std::string gps_serial() const;

   /// CRC of a slave
   uint32_t CRCSlave (int node) const;
    
};


/// Tree strcuture of master/fanout and slave modules
class mfo_tree : public mfo_t
{
public:
   /// Default constructor
   mfo_tree();
   /// Donstructor
   ~mfo_tree();
   /// Copy constructor
   mfo_tree(const mfo_tree& mfo);
   /// Copy constructor
   explicit mfo_tree(const mfo_t& mfo, const std::string& name = "");
   /// Copy operator
   mfo_tree& operator= (const mfo_tree& mfo);
   /// Copy operator
   mfo_tree& operator= (const mfo_t& mfo);
   
   /// Is root node?
   bool is_root() const;
   /// Downstream node is a fanout and active
   bool is_fanout (int node) const;
   /// Returns parent node
   mfo_tree* parent();
   /// Returns parent node
   const mfo_tree* parent() const;
   /// Returns downstream fanout node (fanout is not released)
   mfo_tree* fanout(int node);
   /// Returns downstream fanout node (fanout is not released)
   const mfo_tree* fanout(int node) const;
   /// Set parent node
   void set_parent (mfo_tree* parent);
   /// Set fanout node
   bool set_fanout (int node, mfo_tree* fanout);
   /// Set fanout node (fanout is copied)
   bool set_fanout (int node, const mfo_tree& fanout);
   /// Returns the node of fanout if it is in the slave list, or -1 if not
   int fanout_node (const mfo_tree& mfo) const;
   /// Set fanout node if mfo is part of the slave tree (mfo is copied)
   bool set_fanout (const mfo_tree& mfo);
   
   /// Name of module
   std::string		mName;

protected:

   /// parent node pointer (not owned by class)
   mfo_tree*		mParent;
   /// fanout node pointer (owned by class)
   mfo_tree*		mFanout[16];
};

typedef std::list<mfo_tree> mfo_tree_list;


/** xsilMfoPort converts a mfo_fanout_t object to an output stream in LIGO_LW
    format as described in LIGO-T000067.
    @memo Writes mfo_fanout_t in LIGO_LW format
    @version 0.1
  *************************************************************************/
class xsilMfoPort {
protected:
   /// Indent level
   int            fLevel;
   /// Pointer to mfo_tree object
   const mfo_fanout_t *fData;
   /// Pointer to name
   const char* 	  fName;
public:
   /// Default constructor
   xsilMfoPort (int level = 1) : fLevel(level), fData (0), fName (0) {
   }
   /// Constructor for xsilMfoPort
   xsilMfoPort(const mfo_fanout_t *port, const char* name = 0, int level = 1):
   fLevel(level),fData(port),fName(name) {
   }
   /// Constructor for xsilMfoPort
   xsilMfoPort(const mfo_fanout_t& port, const char* name = 0, int level = 1):
   fLevel(level),fData(&port),fName(name) {
   }
   /// Write the data to output stream
   std::ostream& write (std::ostream &os) const;
};

/** Write mfo_fanout_t to an output stream.
    @param os output stream
    @param port xsilMfoPort object
    @memo XML output of mfo_fanout_t
    @return output stream
  *************************************************************************/
inline std::ostream& operator<< (std::ostream &os, const xsilMfoPort &port) {
      return port.write(os); }



/** xsilSlaveBasic converts a slave_basic_t object to an output stream in 
    LIGO_LW format as described in LIGO-T000067. Used by xsilSlave.
    @memo Writes slave_basic_t in LIGO_LW format
    @version 0.1
  *************************************************************************/
class xsilSlaveBasic {
protected:
   /// Indent level
   int            fLevel;
   /// Pointer to mfo_tree object
   const slave_basic_t *fData;
public:
   /// Deafult constructor
   xsilSlaveBasic (int level = 1) : fLevel(level), fData (0) {
   }
   /// Constructor
   xsilSlaveBasic(const slave_basic_t *slave, int level = 1):
   fLevel(level),fData(slave) {
   }
   /// Constructor
   xsilSlaveBasic(const slave_basic_t& slave, int level = 1):
   fLevel(level),fData(&slave) {
   }
   /// Write the data to output stream
   std::ostream& write (std::ostream &os) const;
};

/** Write slave_basic_t to an output stream.
    @param os output stream
    @param slave xsilSlaveBasic object
    @memo XML output of slave_basic_t
    @return output stream
  *************************************************************************/
inline std::ostream& operator<< (std::ostream &os, 
	                         const xsilSlaveBasic &slave) {
      return slave.write(os); }


/** xsilSlave converts a slave_t object to an output stream in LIGO_LW
    format as described in LIGO-T000067.
    @memo Writes slave_t in LIGO_LW format
    @version 0.1
  *************************************************************************/
class xsilSlave {
protected:
   /// Indent level
   int            fLevel;
   /// Pointer to mfo_tree object
   const slave_t *fData;
   /// Pointer to name
   const char* 	  fName;
public:
   /// Default constructor
   xsilSlave (int level = 1) : fLevel(level), fData (0), fName (0) {
   }
   /// Constructor
   xsilSlave(const slave_t *slave, const char* name = 0, int level = 1):
   fLevel(level),fData(slave),fName(name) {
   }
   /// Constructor
   xsilSlave(const slave_t& slave, const char* name = 0, int level = 1):
   fLevel(level),fData(&slave),fName(name) {
   }
   /// Write the data to output stream
   std::ostream& write (std::ostream &os) const;
};

/** Write slave_t to an output stream.
    @param os output stream
    @param slave xsilSlave object
    @memo XML output of slave_t
    @return output stream
  *************************************************************************/
inline std::ostream& operator<< (std::ostream &os, const xsilSlave &slave) {
      return slave.write(os); }


/** xsilSlaveFanout converts a slave_fanout_t object to an output 
    stream in LIGO_LW format as described in LIGO-T000067. Used by 
    xsilSlave.
    @memo Writes slave_fanout_t in LIGO_LW format
    @version 0.1
  *************************************************************************/
class xsilSlaveFanout {
protected:
   /// Indent level
   int            fLevel;
   /// Pointer to mfo_tree object
   const slave_fanout_t *fData;
public:
   /// Default constructor
   xsilSlaveFanout (int level = 1) : fLevel(level), fData (0) {
   }
   /// Constructor
   xsilSlaveFanout (const slave_t *slave, int level = 1):
   fLevel(level),fData((slave_fanout_t*)slave) {
   }
   /// Constructor
   xsilSlaveFanout (const slave_t& slave, int level = 1):
   fLevel(level),fData((slave_fanout_t*)&slave) {
   }
   /// Write the data to output stream
   std::ostream& write (std::ostream &os) const;
};

/** Write slave_fanout_t to an output stream.
    @param os output stream
    @param slave xsilSlaveFanout object
    @memo XML output of slave_fanout_t
    @return output stream
  *************************************************************************/
inline std::ostream& operator<< (std::ostream &os, 
	                         const xsilSlaveFanout &slave) {
      return slave.write(os); }


/** xsilSlaveDuoTone converts a slave_duotone_t object to an output 
    stream in LIGO_LW format as described in LIGO-T000067. Used by 
    xsilSlave.
    @memo Writes slave_duotone_t in LIGO_LW format
    @version 0.1
  *************************************************************************/
class xsilSlaveDuoTone {
protected:
   /// Indent level
   int            fLevel;
   /// Pointer to mfo_tree object
   const slave_duotone_t *fData;
public:
   /// Default constructor
   xsilSlaveDuoTone (int level = 1) : fLevel(level), fData (0) {
   }
   /// Constructor
   xsilSlaveDuoTone (const slave_t *slave, int level = 1):
   fLevel(level),fData((slave_duotone_t*)slave) {
   }
   /// Constructor
   xsilSlaveDuoTone (const slave_t& slave, int level = 1):
   fLevel(level),fData((slave_duotone_t*)&slave) {
   }
   /// Write the data to output stream
   std::ostream& write (std::ostream &os) const;
};

/** Write slave_duotone_t to an output stream.
    @param os output stream
    @param slave xsilSlaveDuoTone object
    @memo XML output of slave_duotone_t
    @return output stream
  *************************************************************************/
inline std::ostream& operator<< (std::ostream &os, 
	                         const xsilSlaveDuoTone &slave) {
      return slave.write(os); }


/** xsilSlaveComparator converts a slave_comparator_t object to an output 
    stream in LIGO_LW format as described in LIGO-T000067. Used by 
    xsilSlave.
    @memo Writes slave_comparator_t in LIGO_LW format
    @version 0.1
  *************************************************************************/
class xsilSlaveComparator {
protected:
   /// Indent level
   int            fLevel;
   /// Pointer to mfo_tree object
   const slave_comparator_t *fData;
public:
   /// Default constructor
   xsilSlaveComparator (int level = 1) : fLevel(level), fData (0) {
   }
   /// Constructor
   xsilSlaveComparator(const slave_t *slave, int level = 1):
   fLevel(level),fData((const slave_comparator_t*)slave) {
   }
   /// Constructor
   xsilSlaveComparator(const slave_t& slave, int level = 1):
   fLevel(level),fData((const slave_comparator_t*)&slave) {
   }
   /// Write the data to output stream
   std::ostream& write (std::ostream &os) const;
};

/** Write slave_comparator_t to an output stream.
    @param os output stream
    @param slave xsilSlaveComparator object
    @memo XML output of slave_comparator_t
    @return output stream
  *************************************************************************/
inline std::ostream& operator<< (std::ostream &os, 
	                         const xsilSlaveComparator &slave) {
      return slave.write(os); }


/** xsilSlaveXOLocking converts a slave_xolocking_t object to an output 
    stream in LIGO_LW format as described in LIGO-T000067. Used by 
    xsilSlave.
    @memo Writes slave_xolocking_t in LIGO_LW format
    @version 0.1
  *************************************************************************/
class xsilSlaveXOLocking {
protected:
   /// Indent level
   int            fLevel;
   /// Pointer to mfo_tree object
   const slave_xolocking_t *fData;
public:
   /// Default constructor
   xsilSlaveXOLocking (int level = 1) : fLevel(level), fData (0) {
   }
   /// Constructor
   xsilSlaveXOLocking(const slave_t *slave, int level = 1):
   fLevel(level),fData((slave_xolocking_t*)slave) {
   }
   /// Constructor
   xsilSlaveXOLocking(const slave_t& slave, int level = 1):
   fLevel(level),fData((slave_xolocking_t*)&slave) {
   }
   /// Write the data to output stream
   std::ostream& write (std::ostream &os) const;
};

/** Write slave_xolocking_t to an output stream.
    @param os output stream
    @param slave xsilSlaveXOLocking object
    @memo XML output of slave_xolocking_t
    @return output stream
  *************************************************************************/
inline std::ostream& operator<< (std::ostream &os, 
	                         const xsilSlaveXOLocking &slave) {
      return slave.write(os); }


/** xsilMfoTree converts a mfo_tree object to an output stream in LIGO_LW
    format as described in LIGO-T000067.
    @memo Writes mfo_tree in LIGO_LW format as described in LIGO-TOOOO67
    @version 0.1
  *************************************************************************/
class xsilMfoTree {
protected:
   /// Indent level
   int            fLevel;
   /// Pointer to mfo_tree object
   const mfo_tree *fData;
   /// Pointer to name
   const char* 	  fName;
public:
   /// Default constructor
   xsilMfoTree (int level = 1) : fLevel(level), fData (0), fName (0) {
   }
   /// Constructor 
   xsilMfoTree(const mfo_tree *mfo, const char* name = 0, int level = 1):
   fLevel(level),fData(mfo),fName(name) {
   }
   /// Constructor 
   xsilMfoTree(const mfo_tree& mfo, const char* name = 0, int level = 1):
   fLevel(level),fData(&mfo),fName(name) {
   }
   /// Write the data to output stream
   std::ostream& write (std::ostream &os) const;
};

/** Write mfo_tree to an output stream.
    @param os output stream
    @param mfo mfo_tree_list object
    @memo XML output of mfo_tree
    @return output stream
  *************************************************************************/
inline std::ostream& operator<< (std::ostream &os, 
			 	 const xsilMfoTree &mfo) {
      return mfo.write(os); }


/** xsilMfoTreeList converts a list of mfo_tree object to an output stream 
    in LIGO_LW format as described in LIGO-T000067. This includes XML header.
    @memo Writes mfo_tree in LIGO_LW format as described in LIGO-TOOOO67
    @version 0.1
  *************************************************************************/
class xsilMfoTreeList {
protected:
   /// Indent level
   int            fLevel;
   /// Pointer to list of mfo_tree objects
   const mfo_tree_list *fData;
   /// Time
   Time		  fTime;
public:
   /// Default constructor
   xsilMfoTreeList (int level = 1) : fLevel(level), fData (0) {
   }
   /// Constructor 
   xsilMfoTreeList(const mfo_tree_list *mfo, const Time& t, int level = 1):
   fLevel(level),fData(mfo),fTime(t) {
   }
   /// Constructor
   xsilMfoTreeList(const mfo_tree_list& mfo, const Time& t, int level = 1):
   fLevel(level),fData(&mfo),fTime(t) {
   }
   /// Write the data to output stream
   std::ostream& write (std::ostream &os) const;
};

/** Write mfo_tree_list to an output stream.
    @param os output stream
    @param mfo xsilMfoTreeList object
    @memo XML output of mfo_tree_list
    @return output stream
  *************************************************************************/
inline std::ostream& operator<< (std::ostream &os, 
				 const xsilMfoTreeList &mfo) {
      return mfo.write(os); }


/** Handler for mfo_fanout_t data types
    @memo mfo_fanout_t datatype handler
  *************************************************************************/
class xsilHandlerMfoFanout: public xml::xsilHandler {
protected:
   /// Pointer to list of mfo_tree classes
   mfo_fanout_t*                 fObject;
   
   public:
   /// Constructor
   explicit xsilHandlerMfoFanout (mfo_fanout_t& object, 
                        const attrlist *attr=0, bool ignore=false);
   /// Destructor
   virtual ~xsilHandlerMfoFanout();
   /// bool parameter callback (must return true if handled)
   virtual bool HandleParameter (const std::string& name,const attrlist& attr,
                        const bool& p, int N = 1);
   /// byte parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const char& p, int N = 1) {
      return false; }
   /// short parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const short& p, int N = 1) {
      return false; }
   /// int parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const int& p, int N=1);
#ifndef __CINT__
   /// long parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, 
                        const attrlist& attr,
                        const long long& p, int N=1) {
      return false; }
#endif //__CINT__
   /// float parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, 
                        const attrlist& attr, 
                        const float& p, int N=1) {
      return false; }
   /// double parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, 
                        const attrlist& attr, 
                        const double& p, int N=1);
   /// complex float parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const std::complex<float>& p, int N = 1) {
      return false; }
   /// complex double parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const std::complex<double>& p, int N = 1) {
      return false; }
   /// string parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const std::string& p) {
      return false; }
   /// time callback (must return true if handled)
   virtual bool HandleTime (const std::string& name, const attrlist& attr,
                        unsigned long sec, unsigned long nsec) {
      return false; }
   /// data callback (must return true if data is adopted)
   virtual bool HandleData (const std::string& name, float* x, 
                        int dim1, int dim2=0, int dim3=0, int dim4=0) {
      return false; }
   /// data callback (must return true if data is adopted)
   virtual bool HandleData (const std::string& name, std::complex<float>* x,
                        int dim1, int dim2=0, int dim3=0, int dim4=0) {
      return false; }			
   /// data callback (must return true if data is adopted)
   virtual bool HandleData (const std::string& name, double* x, 
                        int dim1, int dim2=0, int dim3=0, int dim4=0) {
      return false; }
   /// data callback (must return true if data is adopted)
   virtual bool HandleData (const std::string& name, std::complex<double>* x,
                        int dim1, int dim2=0, int dim3=0, int dim4=0) {
      return false; }
   /// handler for nested data objects
   virtual xsilHandler* GetHandler (const attrlist& attr) {
      return false; }
};

/** Handler for slave_t data types
    @memo slave_t datatype handler
  *************************************************************************/
class xsilHandlerSlave: public xml::xsilHandler {
protected:
   /// Pointer to list of mfo_tree classes
   slave_t*                    fObject;
   
   public:
   /// Constructor
   explicit xsilHandlerSlave (slave_t& object, 
                        const attrlist *attr=0, bool ignore=false);
   /// Destructor
   virtual ~xsilHandlerSlave();
   /// bool parameter callback (must return true if handled)
   virtual bool HandleParameter (const std::string& name,const attrlist& attr,
                        const bool& p, int N = 1);
   /// byte parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const char& p, int N = 1) {
      return false; }
   /// short parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const short& p, int N = 1) {
      return false; }
   /// int parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const int& p, int N=1);
#ifndef __CINT__
   /// long parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, 
                        const attrlist& attr,
                        const long long& p, int N=1) {
      return false; }
#endif //__CINT__
   /// float parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, 
                        const attrlist& attr, 
                        const float& p, int N=1) {
      return false; }
   /// double parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, 
                        const attrlist& attr, 
                        const double& p, int N=1);
   /// complex float parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const std::complex<float>& p, int N = 1) {
      return false; }
   /// complex double parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const std::complex<double>& p, int N = 1) {
      return false; }
   /// string parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const std::string& p) {
      return false; }
   /// time callback (must return true if handled)
   virtual bool HandleTime (const std::string& name, const attrlist& attr,
                        unsigned long sec, unsigned long nsec);
   /// data callback (must return true if data is adopted)
   virtual bool HandleData (const std::string& name, float* x, 
                        int dim1, int dim2=0, int dim3=0, int dim4=0) {
      return false; }
   /// data callback (must return true if data is adopted)
   virtual bool HandleData (const std::string& name, std::complex<float>* x,
                        int dim1, int dim2=0, int dim3=0, int dim4=0) {
      return false; }			
   /// data callback (must return true if data is adopted)
   virtual bool HandleData (const std::string& name, double* x, 
                        int dim1, int dim2=0, int dim3=0, int dim4=0) {
      return false; }
   /// data callback (must return true if data is adopted)
   virtual bool HandleData (const std::string& name, std::complex<double>* x,
                        int dim1, int dim2=0, int dim3=0, int dim4=0) {
      return false; }
   /// handler for nested data objects
   virtual xsilHandler* GetHandler (const attrlist& attr) {
      return false; }
};

/** Handler for mfo_tree data types
    @memo mfo_tree datatype handler
  *************************************************************************/
class xsilHandlerMfoTree: public xml::xsilHandler {
protected:
   /// Pointer to list of mfo_tree classes
   mfo_tree*                      fObject;
   
   public:
   /// Constructor
   explicit xsilHandlerMfoTree (mfo_tree& object, 
                     const attrlist *attr=0, bool ignore=false);
   /// Destructor
   virtual ~xsilHandlerMfoTree();
   /// bool parameter callback (must return true if handled)
   virtual bool HandleParameter (const std::string& name,const attrlist& attr,
                        const bool& p, int N = 1);
   /// byte parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const char& p, int N = 1) {
      return false; }
   /// short parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const short& p, int N = 1) {
      return false; }
   /// int parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const int& p, int N=1);
#ifndef __CINT__
   /// long parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, 
                        const attrlist& attr,
                        const long long& p, int N=1) {
      return false; }
#endif //__CINT__
   /// float parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, 
                        const attrlist& attr, 
                        const float& p, int N=1) {
      return false; }
   /// double parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, 
                        const attrlist& attr, 
                        const double& p, int N=1);
   /// complex float parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const std::complex<float>& p, int N = 1) {
      return false; }
   /// complex double parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const std::complex<double>& p, int N = 1) {
      return false; }
   /// string parameter callback (must return true if handled)
   virtual bool HandleParameter(const std::string& name, const attrlist& attr,
                        const std::string& p);
   /// time callback (must return true if handled)
   virtual bool HandleTime (const std::string& name, const attrlist& attr,
                        unsigned long sec, unsigned long nsec);
   /// data callback (must return true if data is adopted)
   virtual bool HandleData (const std::string& name, float* x, 
                        int dim1, int dim2=0, int dim3=0, int dim4=0) {
      return false; }
   /// data callback (must return true if data is adopted)
   virtual bool HandleData (const std::string& name, std::complex<float>* x,
                        int dim1, int dim2=0, int dim3=0, int dim4=0) {
      return false; }			
   /// data callback (must return true if data is adopted)
   virtual bool HandleData (const std::string& name, double* x, 
                        int dim1, int dim2=0, int dim3=0, int dim4=0) {
      return false; }
   /// data callback (must return true if data is adopted)
   virtual bool HandleData (const std::string& name, std::complex<double>* x,
                        int dim1, int dim2=0, int dim3=0, int dim4=0) {
      return false; }
   /// handler for nested data objects
   virtual xsilHandler* GetHandler (const attrlist& attr);
};

/** Xsil mfo_tree handler query class.
    The query will return a handler if the data object type is 
    TimingMasterFanout.
    @memo Xsil mfo_tree handler query
  *************************************************************************/
class xsilHandlerQueryMfoTree: public xml::xsilHandlerQuery {
protected:
   /// Pointer to list of mfo_tree classes
   mfo_tree_list*                 fObjects;
   
public:
   /// Constructor
   explicit xsilHandlerQueryMfoTree (mfo_tree_list &l)
   : fObjects(&l) {
   }
   /// returns a handler for the specified object (or 0 if not)
   virtual xml::xsilHandler* GetHandler (const attrlist& attr);
};



#endif
